/*
 * Support for executing code in new tasks and threads.
 *
 * Copyright (C) 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/re/l4aux.h>
#include <l4/sys/err.h>
#include <l4/util/elf.h>
#include <l4/util/util.h>

#include <exec/elf.h>
#include <exec/process.h>
#include <fsclient/client.h>
#include <ipc/mem_ipc.h>
#include <ipc/server.h>
#include <mem/memory_utils.h>
#include <systypes/fcntl.h>

#include <map>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pthread-l4.h>
#include <pthread.h>

#include "pager_object_interface.h"
#include "pager_object_server.h"



/* A simple system pager also acting as a region mapper. */

typedef std::map<l4_addr_t, MappedRegion> MappedRegions;

class ExecPager : public PagerObject
{
protected:
  MappedRegions _regions;

public:
  virtual void add(MappedRegion region)
  {
    _regions[region.map_start] = region;
  }

  /* Notification methods. */

  virtual long exception(l4_exc_regs_t regs,
                         l4_snd_fpage_t *region);

  virtual long page_fault(l4_umword_t pfa, l4_umword_t pc,
                          l4_snd_fpage_t *region);

  /* Region manager/mapper methods. */

  virtual long attach(address_t *start, offset_t size, map_flags_t flags,
                      l4_cap_idx_t ds, address_t offset, unsigned char align);

};

/* Handle a general exception. */

long ExecPager::exception(l4_exc_regs_t regs, l4_snd_fpage_t *region)
{
  (void) region;

  printf("exception(...) -> pfa = %lx, pc = %lx\n", l4_utcb_exc_pfa(&regs), l4_utcb_exc_pc(&regs));

  printf("r15 = %lx\n", regs.r15);
  printf("r14 = %lx\n", regs.r14);
  printf("r13 = %lx\n", regs.r13);
  printf("r12 = %lx\n", regs.r12);
  printf("r11 = %lx\n", regs.r11);
  printf("r10 = %lx\n", regs.r10);
  printf("r9 = %lx\n", regs.r9);
  printf("r8 = %lx\n", regs.r8);
  printf("rdi = %lx\n", regs.rdi);
  printf("rsi = %lx\n", regs.rsi);
  printf("rbp = %lx\n", regs.rbp);
  printf("pfa = %lx\n", regs.pfa);
  printf("rbx = %lx\n", regs.rbx);
  printf("rdx = %lx\n", regs.rdx);
  printf("rcx = %lx\n", regs.rcx);
  printf("rax = %lx\n", regs.rax);
  printf("trapno = %lx\n", regs.trapno);
  printf("err = %lx\n", regs.err);
  printf("ip = %lx\n", regs.ip);
  printf("flags = %lx\n", regs.flags);
  printf("sp = %lx\n", regs.sp);
  printf("ss = %lx\n", regs.ss);
  printf("fs_base = %lx\n", regs.fs_base);
  printf("gs_base = %lx\n", regs.gs_base);

  return L4_EOK;
}

#define DEBUG 0

/* Handle a page fault using any configured regions. */

long ExecPager::page_fault(l4_umword_t pfa, l4_umword_t pc, l4_snd_fpage_t *region)
{
  l4_umword_t addr = pfa & ~7UL, flags = pfa & 7;

#if DEBUG
  printf("page_fault(%lx, %lx) -> %lx (%lx) -> ", pfa, pc, addr, flags);
#endif

  MappedRegions::iterator it = _regions.upper_bound(addr);

  if (it != _regions.begin())
    it--;
  else
  {
    printf("not mapped!\n");
    return -L4_ENOMEM;
  }

  MappedRegion &r = it->second;

  if ((addr >= r.map_start) && (addr < r.map_start + r.size))
  {
    l4_addr_t page_addr = trunc(addr, L4_PAGESIZE);

    region->fpage = l4_fpage(r.start + (page_addr - r.map_start), L4_PAGESHIFT, r.flags);
    region->snd_base = page_addr;

#if DEBUG
    printf("%lx...%lx from %lx...%lx offset %lx size %d rights %x\n",
           r.map_start, region->snd_base,
           r.start, l4_fpage_memaddr(region->fpage),
           addr - r.map_start,
           l4_fpage_size(region->fpage),
           l4_fpage_rights(region->fpage));
    printf("%lx -> ", addr);

    for (unsigned int i = 0; i < sizeof(l4_umword_t); i++)
      printf("%02x", *((unsigned char *)(r.start + (addr - r.map_start) + i)));

    printf("\n");
#endif

    if (r.flags & L4RE_RM_F_W)
      l4_touch_rw((const void *) (r.start + (page_addr - r.map_start)), L4_PAGESIZE);
    else
      l4_touch_ro((const void *) (r.start + (page_addr - r.map_start)), L4_PAGESIZE);

    return L4_EOK;
  }

#if DEBUG
  printf("not mapped!\n");
#endif

  return -L4_ENOMEM;
}

/* Attach a region for provision when page faults occur. This is required in
   the initialisation of a program by the C library which requires a region
   mapper. */

long ExecPager::attach(address_t *start, offset_t size, map_flags_t flags,
                       l4_cap_idx_t ds, address_t offset, unsigned char align)
{
#if DEBUG
  printf("attach(%lx, %ld, %lx, ..., %lx, %d)\n", *start, size, flags, offset, align);
#endif

  if (align < L4_PAGESHIFT)
    align = L4_PAGESHIFT;

  offset_t increment = 1UL << align;
  offset_t region_size = round(size, increment);

  /* Either attempt to find an address for the specified region, starting from
     any indicated address. */

  if (flags & L4RE_RM_F_SEARCH_ADDR)
  {
    address_t region_start = trunc(*start, increment);
    MappedRegions::iterator it = _regions.upper_bound(*start);

    if (!region_start)
      region_start += increment;

#if DEBUG
    printf("-> search from %lx -> %lx...\n", *start, region_start);
#endif

    /* Before last known region. */

    while (it != _regions.end())
    {
      MappedRegions::iterator next = it;
      MappedRegion &r = it->second;
      address_t start_limit;
      address_t end_limit = r.map_start;

      /* Consider any preceding region. If no such region exists, choose an
         address at the start of memory. */

      if (it == _regions.begin())
        start_limit = L4_PAGESIZE;
      else
      {
        it--;
        MappedRegion &pr = it->second;
        start_limit = pr.map_start + pr.size;
        it = next;
      }

      /* Test against the limits. */

      if (region_start < start_limit)
        region_start = round(start_limit, increment);

      /* Investigate subsequent regions if not enough space exists between the
         preceding region (or start of memory) and the current region. */

      if ((region_start + region_size) > end_limit)
      {
        it++;
        if (it == _regions.end())
          return -L4_ENOMEM;
      }
      else
        break;
    }

    /* Attach the provided dataspace.
       NOTE: This is only done in this implementation to support the paging
             mechanism. In a region mapper residing within the actual task, the
             dataspace's map operation would be invoked to obtain mappings. */

    l4_addr_t ds_start;

    long err = ipc_attach_dataspace(ds, size, (void **) &ds_start);

    if (err)
      return err;

    l4_touch_rw((const void *) ds_start, size);

#if DEBUG
    printf("-> added region for %lx size %ld (%d)\n", region_start, region_size, page_order(region_size));
#endif

    add(MappedRegion(ds_start, region_size, flags & L4RE_DS_F_RIGHTS_MASK, region_start));

    *start = region_start;
    return L4_EOK;
  }

  /* Or attempt to add the specified region at a specific address. */

  else
  {
    // NOTE: To be implemented.

    return -L4_ENOMEM;
  }
}



static ExecPager exec_pager;

static void init_pager(ipc_server_config_type *config)
{
  ipc_server_init_config(config);

  config->expected_items = PagerObject_expected_items;
  config->handler = (ipc_server_handler_type) handle_PagerObject;
  config->handler_obj = static_cast<PagerObject *>(&exec_pager);
}

static long start_pager(ipc_server_config_type *config, pthread_t thread)
{
  config->config_thread = 1;
  config->thread = pthread_l4_cap(thread);

  printf("Starting pager thread...\n");
  return ipc_server_start_config(config);
}



int main(int argc, char *argv[])
{
  long err;

  if (argc < 2)
  {
    printf("Need a program to run.\n");
    return 1;
  }

  /* Obtain the payload as a dataspace. */

  file_t *file = client_open(argv[1], O_RDONLY);

  if (file == NULL)
  {
    printf("Could not read file: %s\n", argv[1]);
    return 1;
  }

  /* Obtain metadata from the file. */

  char buf[4096];
  offset_t nread;

  nread = client_read(file, buf, EI_NIDENT);
  if ((nread < EI_NIDENT) || memcmp(buf, "\x7f" "ELF", 4))
  {
    printf("Not an ELF payload: %s\n", argv[1]);
    return 1;
  }

  Payload *payload = get_payload(buf);

  if (payload == NULL)
  {
    printf("Unrecognised object size.\n");
    return 1;
  }

  client_seek(file, 0, SEEK_SET);
  nread = client_read(file, buf, payload->header_extent());

  if (nread < payload->header_extent())
  {
    printf("Header incomplete.\n");
    return 1;
  }

  printf("Program start: %lx\n", payload->entry_point());

  client_seek(file, 0, SEEK_SET);
  nread = client_read(file, buf, payload->program_header_extent());

  if (nread < payload->program_header_extent())
  {
    printf("Program headers incomplete.\n");
    return 1;
  }

  /* Make appropriate segments, although program segments could be made
     interoperable with these. */

  Segment *segments[payload->segments() + 1];

  for (unsigned int i = 0; i < payload->segments(); i++)
  {
    ProgramSegment *ps = payload->segment(i);

    printf("Segment(0x%lx, 0x%lx, 0x%x, 0x%lx, 0x%lx): %s\n",
           ps->region_address(),
           ps->region_size(),
           ps->region_flags(),
           ps->file_offset(),
           ps->file_contents(),
           ps->loadable() ? "loadable" : "other");

    if (ps->loadable())
      segments[i] = ps->segment();
    else
      segments[i] = NULL;
  }

  /* Copy the payload regions to new dataspaces. */

  address_t program_start = payload->entry_point();
  offset_t initial_stack_size = 16 * L4_PAGESIZE;

  Segment stack(Utcb_area_start - initial_stack_size, initial_stack_size, L4_FPAGE_RW);

  segments[payload->segments()] = &stack;

  for (unsigned int i = 0; i < payload->segments() + 1; i++)
  {
    Segment *segment = segments[i];

    if (segment == NULL)
      continue;

    if (segment->file_contents())
    {
      flags_t flags;

      if ((segment->region_flags() & L4RE_RM_F_RW) == L4RE_RM_F_RW)
        flags = O_RDWR;
      else if (segment->region_flags() & L4RE_RM_F_W)
        flags = O_WRONLY;
      else
        flags = O_RDONLY;

      file_t *file = client_open(argv[1], flags);

      if (file == NULL)
      {
        printf("Could not open file for segment.\n");
        return 1;
      }

      err = segment->fill(file);

      if (err)
      {
        printf("Could not fill segment from file.\n");
        return 1;
      }
    }
    else
    {
      err = segment->allocate();

      if (err)
      {
        printf("Could not allocate segment.\n");
        return 1;
      }
    }
  }

  /* Start the pager. */

  ipc_server_config_type config;
  pthread_t pager_thread;
  pthread_attr_t attr;

  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

  init_pager(&config);

  for (unsigned int i = 0; i < payload->segments() + 1; i++)
  {
    Segment *segment = segments[i];

    if (segment == NULL)
      continue;

    exec_pager.add(segment->region());
  }

  err = pthread_create(&pager_thread, &attr, ipc_server_start_mainloop, &config);

  if (err)
  {
    printf("Could not start pager thread.\n");
    return 1;
  }

  err = start_pager(&config, pager_thread);

  if (err)
  {
    printf("Could not start pager.\n");
    return 1;
  }

  /* Configure the environment for the task, specifying the pager (and exception
     handler plus region mapper). */

  Process process;

  err = process.configure(config.server);

  if (err)
  {
    printf("Could not configure task.\n");
    return 1;
  }

  /* Populate a thread stack with argument and environment details. */

  Stack st(stack);

  /* NOTE: Environment vector is currently not defined. */

  char *envp[] = {NULL};

  st.populate(argc - 1, argv + 1, envp);

  /* Start the new thread in the given stack. */

  printf("Run thread...\n");

  err = process.thread_start(program_start, st);

  if (err)
  {
    printf("Could not run thread.\n");
    return 1;
  }

  printf("Finished.\n");
  while (1);

  return 0;
}

/* vim: tabstop=2 expandtab shiftwidth=2
*/
